"use strict";

exports.isFuture = isFuture;

exports.Scheduler = Scheduler;

exports.Future = Future;

exports.isDeinitError = isDeinitError;

function isFuture(value) {
    return isObject(value) && isFunction(value.settle) && isFunction(value.map) && isFunction(value.deinit);
}

function Queue() {
    var self = this;
    validateInstance(self, Queue);
    self.a = [];
    self.b = [];
    self.c = 0;
    self.d = 0;
    self.length = 0;
}

Queue.prototype = {
    constructor: Queue,
    push: function(value) {
        this.length++;
        this.b.push(value);
    },
    shift: function() {
        var self = this;
        if (self.c >= self.d) {
            var tail = self.a;
            tail.length = 0;
            self.a = self.b;
            self.b = tail;
            self.c = 0;
            self.d = self.a.length;
            if (!self.d) return undefined;
        }
        var value = self.a[self.c];
        self.a[self.c++] = undefined;
        self.length--;
        return value;
    }
};

function Scheduler(deque) {
    var self = this;
    validateInstance(self, Scheduler);
    validate(deque, isFunction);
    self.e = new Queue();
    self.f = deque;
    self.g = false;
    self.h = scheduledTick.bind(null, self);
    self.asap = chooseAsapImplementation(self.h);
}

Scheduler.prototype = {
    constructor: Scheduler,
    push: function(value) {
        this.e.push(value);
        if (this.e.length) scheduleTick(this);
    },
    tick: function() {
        var deque = this.f;
        try {
            while (this.e.length) {
                deque(this.e.shift());
            }
        } finally {
            if (this.e.length) scheduleTick(this);
        }
    },
    deinit: function() {
        while (this.e.length) {
            this.e.shift();
        }
    }
};

function scheduleTick(scheduler) {
    if (!scheduler.g) {
        scheduler.g = true;
        scheduler.asap.call(undefined, scheduler.h);
    }
}

function scheduledTick(scheduler) {
    scheduler.g = false;
    scheduler.tick();
}

var PENDING = 1 | 0;

var ERROR = 2 | 0;

var SUCCESS = 4 | 0;

var PENDING_REJECTION = 8 | 0;

var CONSUMED = 16 | 0;

var MAPPING = 32 | 0;

var SET_ALL = 255 | 0;

var UNSET_STATE = SET_ALL ^ (PENDING | ERROR | SUCCESS);

var SETTLED = ERROR | SUCCESS;

function Future() {
    var self = this;
    validateInstance(self, Future);
    self.i = PENDING;
    self.j = undefined;
    self.k = undefined;
    self.l = undefined;
    self.m = undefined;
    self.n = undefined;
    self.o = undefined;
}

Future.prototype = {
    constructor: Future,
    get state() {
        var self = this;
        return {
            PENDING: someBitsSet(self, PENDING),
            ERROR: someBitsSet(self, ERROR),
            SUCCESS: someBitsSet(self, SUCCESS),
            PENDING_REJECTION: someBitsSet(self, PENDING_REJECTION),
            CONSUMED: someBitsSet(self, CONSUMED),
            MAPPING: someBitsSet(self, MAPPING)
        };
    },
    deref: function() {
        if (someBitsSet(this, ERROR)) {
            unsetBits(this, PENDING_REJECTION);
            throw this.j;
        }
        return this.j;
    },
    settle: function(error, result) {
        var self = this;
        if (someBitsSet(self, SETTLED | MAPPING)) return;
        if (error === self || result === self) {
            throw Error("A future can't be chained to itself");
        }
        if (error) result = undefined; else error = undefined;
        if (isFuture(error)) {
            if (isBaseFuture(error)) {
                if (someBitsSet(error, CONSUMED)) throw Error(CONSUMABLE_ERROR);
                if (!error.l && !self.n) {
                    self.n = alwaysThrow;
                    setupPredecessorSuccessorPair(error, self);
                    return;
                }
            }
            self.k = error.map(function errorSettle(error, result) {
                self.settle(error || result);
            });
            return;
        }
        if (isFuture(result)) {
            if (isBaseFuture(result)) {
                if (someBitsSet(result, CONSUMED)) throw Error(CONSUMABLE_ERROR);
                if (!result.l) {
                    setupPredecessorSuccessorPair(result, self);
                    return;
                }
            }
            self.k = result.map(function valueSettle(error, result) {
                self.settle(error, result);
            });
            return;
        }
        var mapper = self.n;
        self.n = undefined;
        if (mapper) {
            setBits(self, MAPPING);
            try {
                result = mapper(error, result);
                error = undefined;
            } catch (err) {
                error = err;
                result = undefined;
            } finally {
                unsetBits(self, MAPPING);
            }
            self.settle(error, result);
            return;
        }
        self.j = error || result;
        replaceStateBits(self, error ? ERROR : SUCCESS);
        if (error) setBits(self, PENDING_REJECTION);
        scheduleFuture(self);
    },
    map: function(mapper) {
        validate(mapper, isFunction);
        if (someBitsSet(this, CONSUMED)) throw Error(CONSUMABLE_ERROR);
        setBits(this, CONSUMED);
        unsetBits(this, PENDING_REJECTION);
        var future = new Future();
        future.n = mapper;
        setupPredecessorSuccessorPair(this, future);
        return future;
    },
    mapError: function(mapper) {
        validate(mapper, isFunction);
        return this.map(mapError.bind(null, mapper));
    },
    mapResult: function(mapper) {
        validate(mapper, isFunction);
        return this.map(mapResult.bind(null, mapper));
    },
    finally: function(mapper) {
        validate(mapper, isFunction);
        return this.map(mapFinally.bind(null, mapper));
    },
    weak: function() {
        var future = new Future();
        if (!this.m) this.m = new Queue();
        this.m.push(future);
        scheduleFuture(this);
        return future;
    },
    toPromise: function() {
        var self = this;
        if (someBitsSet(self, CONSUMED)) throw Error(CONSUMABLE_ERROR);
        unsetBits(self, PENDING_REJECTION);
        if (someBitsSet(self, ERROR)) return Promise.reject(self.j);
        if (someBitsSet(self, SUCCESS)) return Promise.resolve(self.j);
        return new Promise(mapFutureToPromise.bind(null, self));
    },
    catch: function(onError) {
        return this.toPromise().catch(onError);
    },
    then: function(onResult, onError) {
        return this.toPromise().then(onResult, onError);
    },
    finishPending: function() {
        var self = this;
        if (someBitsSet(self, SETTLED)) {
            var successor = self.l;
            self.l = undefined;
            if (successor) {
                unsetBits(self, PENDING_REJECTION);
                if (someBitsSet(self, ERROR)) successor.settle(self.j); else successor.settle(undefined, self.j);
            }
            while (self.m && self.m.length) {
                var weak = self.m.shift();
                if (someBitsSet(self, ERROR)) weak.settle(self.j); else weak.settle(undefined, self.j);
            }
            finalize(self);
        }
        if (someBitsSet(self, PENDING_REJECTION)) {
            unsetBits(self, PENDING_REJECTION);
            self.constructor.onUnhandledRejection(self);
        }
    },
    deinit: function() {
        var self = this;
        if (someBitsSet(self, MAPPING)) return;
        var prevPredecessor = self.k;
        self.k = undefined;
        try {
            if (!someBitsSet(self, SETTLED)) self.settle(Error(DEINIT_ERROR));
            unsetBits(self, PENDING_REJECTION);
        } finally {
            try {
                finalize(self);
            } finally {
                try {
                    if (prevPredecessor) prevPredecessor.deinit();
                } finally {
                    var nextPredecessor = self.k;
                    self.k = undefined;
                    if (nextPredecessor) nextPredecessor.deinit();
                }
            }
        }
    }
};

Future.from = function from(error, result) {
    var future = new Future();
    future.settle(error, result);
    return future;
};

Future.fromError = function fromError(error) {
    validate(error, Boolean);
    return this.from(error);
};

Future.fromResult = function fromResult(result) {
    return isFuture(result) ? result : this.from(undefined, result);
};

Future.fromPromise = function fromPromise(promise) {
    validate(promise, isPromise);
    var future = new Future();
    promise.then(future.settle.bind(future, undefined), future.settle.bind(future));
    return future;
};

Future.all = function all(values) {
    validate(values, isArray);
    var future = new Future();
    initAllJuncture(new AllJuncture(future, values));
    return future;
};

Future.race = function race(values) {
    validate(values, isArray);
    var future = new Future();
    initRaceJuncture(new RaceJuncture(future, values));
    return future;
};

Future.onUnhandledRejection = function onUnhandledRejection(future) {
    throw future.j;
};

Future.scheduler = new Scheduler(function finishPending(future) {
    future.finishPending();
});

function AllJuncture(future, values) {
    validateInstance(this, AllJuncture);
    this.p = future;
    this.q = values.slice();
}

AllJuncture.prototype.deinit = function deinit() {
    this.p = undefined;
    var values = this.q;
    this.q = undefined;
    forceDeinitFutures(values);
};

function initAllJuncture(all) {
    var values = all.q;
    var settleAll = settleAllJuncture.bind(null, all);
    for (var i = 0; i < values.length; i += 1) {
        var value = values[i];
        if (!isFuture(value)) continue;
        if (isBaseFuture(value)) {
            if (someBitsSet(value, CONSUMED)) throw Error(CONSUMABLE_ERROR);
            if (someBitsSet(value, ERROR)) {
                settleAll(value);
                return;
            }
            if (someBitsSet(value, SUCCESS)) {
                values[i] = quietlyConsume(value);
                continue;
            }
            if (!value.o) {
                setupFinalizer(value, settleAllJunctureAtIndex.bind(null, all, i));
                continue;
            }
        }
        values[i] = value.map(settleAllJunctureAtIndex.bind(null, all, i));
    }
    if (values.some(isFuture)) {
        all.p.o = all.deinit.bind(all);
    } else {
        settleAll();
    }
}

function settleAllJunctureAtIndex(all, i, error, result) {
    var values = all.q;
    if (!values) return;
    if (error) settleAllJuncture(all, error); else {
        values[i] = result;
        if (!values.some(isFuture)) settleAllJuncture(all);
    }
}

function settleAllJuncture(all, error) {
    var future = all.p;
    all.p = undefined;
    var values = all.q;
    all.q = undefined;
    if (error) {
        future.settle(error);
        forceDeinitFutures(values);
    } else {
        future.settle(undefined, values);
    }
}

function RaceJuncture(future, values) {
    validateInstance(this, RaceJuncture);
    this.p = future;
    this.q = values.slice();
}

RaceJuncture.prototype.deinit = function deinit() {
    this.p = undefined;
    var values = this.q;
    this.q = undefined;
    forceDeinitFutures(values);
};

function initRaceJuncture(race) {
    var values = race.q;
    var settleRace = settleRaceJuncture.bind(null, race);
    for (var i = 0; i < values.length; i += 1) {
        var value = values[i];
        if (!isFuture(value)) {
            settleRace(undefined, value);
            return;
        }
        if (isBaseFuture(value)) {
            if (someBitsSet(value, CONSUMED)) throw Error(CONSUMABLE_ERROR);
            if (someBitsSet(value, ERROR)) {
                settleRace(value);
                return;
            }
            if (someBitsSet(value, SUCCESS)) {
                settleRace(undefined, quietlyConsume(value));
                return;
            }
            if (!value.o) {
                setupFinalizer(value, settleRace);
                continue;
            }
        }
        values[i] = value.map(settleRace);
    }
    if (values.some(isFuture)) {
        race.p.o = race.deinit.bind(race);
    } else {
        settleRace();
    }
}

function settleRaceJuncture(race, error, result) {
    var future = race.p;
    race.p = undefined;
    var values = race.q;
    race.q = undefined;
    if (!future) return;
    future.settle(error, result);
    forceDeinitFutures(values);
}

function isDeinitError(value) {
    return isObject(value) && value instanceof Error && value.message === DEINIT_ERROR;
}

var CONSUMABLE_ERROR = "Expected a consumable future: one that has not been mapped or passed to .settle";

var DEINIT_ERROR = "DEINIT";

function alwaysThrow(error, result) {
    throw error || result;
}

function mapError(fun, error, result) {
    return error ? fun(error) : result;
}

function mapResult(fun, error, result) {
    if (error) throw error;
    return fun(result);
}

function mapFinally(fun, error, result) {
    var finalization = fun(error, result);
    if (isFuture(finalization)) {
        return finalization.mapResult(function resetResult() {
            return result;
        });
    }
    if (error) throw error;
    return result;
}

function mapFutureToPromise(future, resolve, reject) {
    function settleFuturePromise(error, result) {
        unsetBits(future, PENDING_REJECTION);
        if (error) reject(error); else resolve(result);
    }
    if (!future.o) setupFinalizer(future, settleFuturePromise); else future.map(settleFuturePromise);
}

function scheduleFuture(future) {
    (future.constructor.scheduler || Future.scheduler).push(future);
}

function setupPredecessorSuccessorPair(predecessor, successor) {
    predecessor.l = successor;
    successor.k = predecessor;
    if (someBitsSet(predecessor, SETTLED)) scheduleFuture(predecessor);
}

function setupFinalizer(future, fun) {
    future.o = fun;
    if (someBitsSet(future, SETTLED)) scheduleFuture(future);
}

function quietlyConsume(future) {
    setBits(future, CONSUMED);
    return future.j;
}

function finalize(future) {
    var finalizer = future.o;
    future.o = undefined;
    if (isFunction(finalizer)) {
        if (someBitsSet(future, ERROR)) finalizer(future.j); else finalizer(undefined, future.j);
    }
}

function isFunction(value) {
    return typeof value === "function";
}

function isArray(value) {
    return isObject(value) && value instanceof Array;
}

function isObject(value) {
    return value != null && typeof value === "object";
}

function isPromise(value) {
    return isObject(value) && isFunction(value.then);
}

function isBaseFuture(value) {
    return isObject(value) && value.constructor === Future;
}

function validate(value, test) {
    if (!test(value)) throw Error("Expected " + value + " to satisfy test " + test.name);
}

function validateInstance(instance, Class) {
    if (!(instance instanceof Class)) throw Error("Cannot call a class as a function");
}

function forceDeinitFutures(list) {
    if (isArray(list)) forceEach(list, deinitFuture);
}

function deinitFuture(value) {
    if (isFuture(value)) value.deinit();
}

function forceEach(list, fun) {
    var error = undefined;
    for (var i = 0; i < list.length; i += 1) {
        var value = list[i];
        try {
            fun(value);
        } catch (err) {
            error = err;
        }
    }
    if (error) throw error;
}

function someBitsSet(future, bitmask) {
    return (future.i & bitmask) !== 0;
}

function setBits(future, bitmask) {
    future.i |= bitmask;
}

function unsetBits(future, bitmask) {
    future.i &= ~bitmask;
}

function replaceStateBits(future, bitmask) {
    future.i = future.i & UNSET_STATE | bitmask;
}

function chooseAsapImplementation(fun) {
    if (typeof self === "undefined" && typeof process === "object" && process && process.nextTick) {
        return process.nextTick;
    }
    if (typeof MessageChannel === "function") {
        var channel = new MessageChannel();
        channel.port1.onmessage = fun;
        return function asap() {
            channel.port2.postMessage(undefined);
        };
    }
    return setTimeout;
}